# coding: utf-8

metadata = dict(
  __name__        = "audio.index",
  __version__     = "1.0.3",
  __license__     = "MIT License",  
  __author__      = u"Sébastien Boisgérault <Sebastien.Boisgerault@mines-paristech.fr>",
  __url__         = None,
  __summary__     = "Acoustic-Phonetic Index",
  __readme__      = None,
  __classifiers__ = None,
)

globals().update(metadata)

__all__ = metadata.keys()
